#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import re
import argparse

from . import charms
from .mr import Mr


def setup_parser():
    parser = argparse.ArgumentParser(prog='charm update',
                                     description='Update charm_directory with '
                                     'latest from Charm Store')
    parser.add_argument('charm_directory', nargs='?',
                        help='Path to where all charms are stored')
    parser.add_argument('-f', '--fix', action='store_true',
                        help='Attempt to fix charms in charm_directory')

    return parser


def update(charm_dir, fix=False):
    mr = Mr(charm_dir)
    for charm in charms.remote():
        if re.match('^lp:charms\/', charm):
            charm_name = os.path.basename(charm)
            if mr.exists(charm_name) and fix:
                mr.update(charm_name, force=True)
                continue
            mr.add(charm_name, charm)
    try:
        mr.save()
    except Exception as e:
        raise Exception(".mrconfig not saved: %s" % e.strerror)


def main():
    parser = setup_parser()
    args = parser.parse_args()

    print 'Pulling charm list from Launchpad'

    try:
        update(args.charm_directory, args.fix)
    except Exception as e:
        # Got this from http://stackoverflow.com/q/5574702/196832
        print >> sys.stderr, ".mrconfig not saved: ", e
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from . import charms


def main():
    if len(sys.argv) > 1:
        print('usage: list [ --help ]')
        if sys.argv[1] == '--help':
            sys.exit(0)
        else:
            sys.exit(1)

    print "\n".join(charms.remote())


if __name__ == "__main__":
    main()



#!/usr/bin/python

#    Copyright (C) 2011 - 2013  Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import argparse

from bundles import Bundle
from charms import Charm
from cli import parser_defaults


def get_args(args=None):
    parser = argparse.ArgumentParser(
        description='Performs static analysis on charms and bundles')
    parser.add_argument('-n', '--offline', action='store_false',
                        help='Only perform offline proofing')
    parser.add_argument('charm_name', nargs='?', default=os.getcwd(),
                        help='path of charm dir to check. Defaults to PWD')
    parser = parser_defaults(parser)
    args = parser.parse_args(args)

    return args


def proof(path, is_bundle=False, with_remote=True, debug=False):
    path = os.path.abspath(path)
    if not is_bundle:
        try:
            c = Charm(path)
        except:
            try:
                c = Bundle(path, debug)
            except Exception as e:
                return ["FATAL: Not a Bundle or a Charm, can not proof"], 200
    else:
        try:
            c = Bundle(path, debug)
        except Exception as e:
            return ["FATAL: %s" % e.strerror], 200

    lint, err_code = c.proof(with_remote)
    return lint, err_code


def main():
    args = get_args()
    lint, exit_code = proof(args.charm_name, args.bundle, args.offline,
                            args.debug)
    if lint:
        print "\n".join(lint)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse

from . import charms


def setup_parser():
    parser = argparse.ArgumentParser(prog='charm search',
                                     description='Match name against all '
                                     'charms (official and personal) in store')
    parser.add_argument('name', nargs=1, help='Name which to search by')

    return parser


def main():
    parser = setup_parser()
    args = parser.parse_args()
    matches = [c for c in charms.remote() if args.name[0] in c]
    print '\n'.join(matches)



import os
import yaml
import glob
import json

from linter import Linter
from charmworldlib import bundle as cw_bundle


class BundleLinter(Linter):
    def validate(self, contents):
        for name, bdata in contents.items():
            if name == 'envExport':
                self.warn('envExport is the default export name. Please '
                          'use a unique name')

            if 'series' not in bdata and 'inherits' not in bdata:
                self.info("%s: No series defined" % name)

            if 'services' in bdata:
                for svc, sdata in bdata['services'].items():
                    if not 'annotations' in sdata:
                        self.warn('%s: %s: No annotations found, will render '
                                  'bad in GUI' % (name, svc))
            else:
                if 'inherits' not in bdata:
                    self.err("%s: No services defined" % name)
                    return

    def local_proof(self, bundle):
        try:
            data = bundle.bundle_file()
        except:
            raise

        readmes = glob.glob(os.path.join(bundle.bundle_path, 'README*'))
        if len(readmes) < 1:
            self.warn('No readme file found')

        self.validate(data)

    def remote_proof(self, bundle):
        bundles = cw_bundle.Bundles()
        deployer_file = bundle.bundle_file(parse=True)
        proof_output = bundles.proof(deployer_file)

        # Loop through errors for verbose outputing.
        # http://paste.mitechie.com/show/1048/
        if self.debug:
            print json.dumps(proof_output, 2)

        if 'error_messages' in proof_output:
            for message in proof_output['error_messages']:
                self.err(message)


class Bundle(object):
    def __init__(self, path, debug=False):
        self.bundle_path = os.path.abspath(path)
        self.supported_files = ['bundles.yaml', 'bundles.json']
        self.debug = debug
        if not self.is_bundle():
            raise Exception('Not a bundle')

    def is_bundle(self):
        for f in self.supported_files:
            if os.path.isfile(os.path.join(self.bundle_path, f)):
                break
        else:
            return False

        if os.path.isfile(os.path.join(self.bundle_path, 'metadata.yaml')):
            return False

        return True

    def bundle_file(self, parse=True):
        for f in self.supported_files:
            if os.path.isfile(os.path.join(self.bundle_path, f)):
                with open(os.path.join(self.bundle_path, f)) as d:
                    return yaml.safe_load(d.read()) if parse else d.read()

        raise Exception('No bundle.json or bundle.yaml file found')

    def proof(self, remote=True):
        lint = BundleLinter(self.debug)
        lint.local_proof(self)
        if remote:
            lint.remote_proof(self)
        else:
            lint.info('No remote checks performed')

        return lint.lint, lint.exit_code

    def promulgate(self):
        pass



#!/usr/bin/python

# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import argparse

from Cheetah.Template import Template

from cli import parser_defaults
from charms import Charm
from charmworldlib import charm as cwc

TPL_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
ATPL = {'deploy': os.path.join(TPL_DIR, 'tests', 'deploy.tpl'),
        'body': os.path.join(TPL_DIR, 'tests', 'body.tpl'),
        'relate': os.path.join(TPL_DIR, 'tests', 'relate.tpl')}
CHARM_TPL = os.path.join(TPL_DIR, 'charm')


def graph(interface, endpoint, series='precise'):
    matches = {'requires': 'provides', 'provides': 'requires'}
    c = cwc.Charms()
    charms = c.search({matches[endpoint]: interface, 'series': series})

    return charms[0]


def readme(charm_dir, is_bundle=False, debug=False):
    c = Charm(charm_dir)

    if not c.is_charm():
        raise Exception('%s is not a charm' % charm_dir)

    shutil.copy(os.path.join(CHARM_TPL, 'README.ex'), charm_dir)


def tests(charm_dir, is_bundle=False, debug=False):
    c = Charm(charm_dir)

    interfaces = {}
    deploy = []
    relations = []

    if not c.is_charm():
        raise Exception('Not a Charm')

    mdata = c.metadata()

    for rel_type in ['provides', 'requires']:
        if rel_type in mdata:
            interfaces[rel_type] = {}
            for rel, data in mdata[rel_type].iteritems():
                iface = data['interface']
                if iface and iface not in interfaces[rel_type]:
                    r = graph(iface, rel_type)
                    interfaces[rel_type][iface] = r
                    deploy.append(r.url)

                relations.append(['%s:%s' % (mdata['name'], rel), r.name])

    d = Template(file=ATPL['deploy'], searchList=[{'services': deploy}])
    s = Template(file=ATPL['relate'], searchList=[{'relations': relations}])

    t = Template(file=ATPL['body'], searchList=[{'deploy': d, 'relate': s}])

    if not os.path.exists(os.path.join(charm_dir, 'tests')):
        os.mkdir(os.path.join(charm_dir, 'tests'))

    with open(os.path.join(charm_dir, 'tests', '00-autogen'), 'w') as f:
        f.write(str(t))

    if not os.path.exists(os.path.join(charm_dir, 'tests', '00-setup')):
        with open(os.path.join(charm_dir, 'tests', '00-setup'), 'w') as f:
            f.write("""#!/bin/bash

add-apt-repository ppa:juju/stable
apt-get update
apt-get install amulet
""")

    os.chmod(os.path.join(charm_dir, 'tests', '00-autogen'), 0755)
    os.chmod(os.path.join(charm_dir, 'tests', '00-setup'), 0755)


def parser(args=None):
    parser = argparse.ArgumentParser(
        description='Builds portions of a charm or bundle')
    parser.add_argument('subcommand', choices=['tests', 'readme'],
                        help='Which type of generator to run')
    parser = parser_defaults(parser)
    return parser.parse_args(args)


def main():
    a = parser()
    if a.subcommand == 'tests':
        tests(os.getcwd(), is_bundle=a.bundle, debug=a.debug)
    elif a.subcommand == 'readme':
        readme(os.getcwd(), is_bundle=a.bundle, debug=a.debug)
    else:
        raise Exception('No subcommand found')


if __name__ == '__main__':
    main()



# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import ConfigParser

from bzrlib import trace
from bzrlib.bzrdir import BzrDir
from bzrlib.branch import Branch
from bzrlib.plugin import load_plugins
from bzrlib.repository import Repository

load_plugins()
trace.enable_default_logging()  # Provide better error handling


class Mr:
    def __init__(self, directory=None, config=None, mr_compat=True):
        self.directory = directory or os.getcwd()
        self.control_dir = os.path.join(self.directory, '.bzr')
        self.config_file = config or os.path.join(self.directory, '.mrconfig')
        self.mr_compat = mr_compat

        if mr_compat:
            if self.__is_repository():
                self.config = self.__read_cfg()
                self.bzr_dir = Repository.open(self.directory)
            else:
                self.config = ConfigParser.RawConfigParser()
                r = BzrDir.create(self.directory)
                self.bzr_dir = r.create_repository(shared=True)
        else:
            self.config = ConfigParser.RawConfigParser()
            self.bzr_dir = None

    def add(self, name, repository='lp:charms', checkout=False):
        # This isn't a true conversion of Mr, as such it's highly specialized
        # for just Charm Tools. So when you "add" a charm, it's just going
        # to use the charm name to fill in a template. Repository is in there
        # just in case we later add personal branching.
        '''Add a respository to the mrconfig'''
        if not name:
            raise Exception('No name provided')
        if not self.config.has_section(name):
            self.config.add_section(name)

        self.config.set(name, 'checkout', "bzr branch %s %s" %
                        (repository, name))

        if checkout:
            self.checkout(name)

    def checkout(self, name=None):
        '''Checkout either one or all repositories from the mrconfig'''
        if not name:
            for name in self.config.sections():
                charm_remote = self.__get_repository(name)
                self.__checkout(charm_remote,
                                os.path.join(self.directory, name))
        else:
            # Move this, and the charm_* stuff to _checkout? Makes sense
            if not self.config.has_section(name):
                raise Exception('No configuration for %s' % name)

            charm_remote = self.__get_repository(name)
            self.__checkout(charm_remote,
                            os.path.join(self.directory, name))

    def update(self, name=None, force=False):
        '''Update, or checkout, a charm in to directory'''
        if name:
            self.__update(name)
        else:
            for charm in self.config.sections():
                self.__update(charm)

    def remove(self, name=None):
        '''Remove a repository from the mrconfig'''
        if not name:
            raise Exception('No name provided')

        self.config.remove_section(name)

    def list(self):
        '''Return all sections of the mr configuration'''
        return self.config.sections()

    def exists(self, name):
        '''Checks if the configuration already exists for this section'''
        return self.config.has_section(name)

    def save(self):
        '''Save the configuration file to disk'''
        with open(self.config_file, 'w') as mrcfg:
            self.config.write(mrcfg)

    __write_cfg = save

    def __read_cfg(self):
        cfg = ConfigParser.ConfigParser()
        if not self.config_file:
            raise Exception('No .mrconfig specified')
        if os.path.exists(self.config_file):
            cfg.read(self.config_file)
        return cfg

    def __checkout(self, src, to):
        remote = Branch.open(src)
        remote.bzrdir.sprout(to)
        # I wish there was a way to 'close' a RemoteBranch. Sadly,
        # I don't think there is

    def __update(self, name):
        if not os.path.exists(os.path.join(self.directory, name, '.bzr')):
            return self.checkout(name)

        charm_remote = self.__get_repository(name)
        local_branch = Branch.open(os.path.join(self.directory, name))
        remote_branch = Branch.open(charm_remote)
        local_branch.pull(remote_branch)

    def __get_repository(self, name):
        if not self.config.has_section(name):
            raise Exception('No section "%s" configured' % name)

        return self.config.get(name, 'checkout').split(' ')[-2]

    def __is_repository(self):
        try:
            r = Repository.open(self.directory)
        except:
            return False

        return r.is_shared()



#!/usr/bin/python
#
#    review - adds review comment to charm bug
#
#    Copyright (C) 2011  Canonical Ltd.
#    Author: Mark Mims <mark.mims@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import argparse
import logging
import sys

from launchpadlib.launchpad import Launchpad


def parse_options():
    parser = argparse.ArgumentParser(
        description="Review a charm by adding a comment to the corresponding"
        " charm bug.  The review comment can be passed as a --message"
        " arg or via stdin")

    parser.add_argument('bug_id',
                        help='The launchpad bug for the charm being reviewed.')

    parser.add_argument(
        '--message', '-m',
        help='The review text to add as a comment to the bug.')

    parser.add_argument(
        '--yes', '-y', dest="skip_prompt", default=False, action='store_true',
        help='do not prompt me... just do it.')

    parser.add_argument(
        '--verbose', '-v', default=False, action='store_true',
        help='show debug logging?')

    parser.add_argument(
        '--lp-instance', '-t', dest='lp_instance', default='production',
        help="The Launchpad instance to use. Defaults to production, but "
        "staging' or 'qastaging' might be used for testing.")

    return parser.parse_args()


def log_level(verbose):
    if verbose:
        return logging.DEBUG
    else:
        return logging.WARNING


def get_message_from_stdin():
    stream = sys.stdin
    text = stream.read()
    sys.stdin = open('/dev/tty')  # 'reset' stdin to prompt if necessary
    return text


def get_message(message):
    if message:
        return message
    else:
        return get_message_from_stdin()


def prompt_to_continue(bug_id):
    logging.debug("prompting")
    ans = raw_input("Really add this comment to launchpad bug #%s? y/[n] "
                    % bug_id)
    return ans.strip().lower().startswith('y')


def main():
    args = parse_options()
    logging.basicConfig(level=log_level(args.verbose),
                        format='%(levelname)s:%(message)s')

    review_message = get_message(args.message)  # before connecting to lp

    logging.debug('login with %s launchpad:', args.lp_instance)
    lp = Launchpad.login_with('charm-pilot', args.lp_instance)

    bug_id = args.bug_id
    logging.debug('find bug %s:', bug_id)
    bug = lp.bugs[bug_id]

    if bug:
        logging.debug('found bug')

        if args.skip_prompt or prompt_to_continue(bug_id):
            logging.debug('adding comment')
            # TODO check return or catch exception
            bug.newMessage(content=review_message)
        else:
            logging.debug('not adding comment')

    else:
        logging.error("no bug: %s", bug_id)
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python

from launchpadlib.launchpad import Launchpad
from operator import itemgetter
import datetime
import itertools
import argparse


def calculate_age(from_date=None):
    if not from_date:
        return None


def format_as_table(data,
                    keys,
                    header=None,
                    sort_by_key=None,
                    sort_order_reverse=False):
    """Takes a list of dictionaries, formats the data, and returns
    the formatted data as a text table.

    Required Parameters:
        data - Data to process (list of dictionaries). (Type: List)
        keys - List of keys in the dictionary. (Type: List)

    Optional Parameters:
        header - The table header. (Type: List)
        sort_by_key - The key to sort by. (Type: String)
        sort_order_reverse - Default sort order is ascending, if
            True sort order will change to descending. (Type: Boolean)
    """
    # Sort the data if a sort key is specified (default sort order
    # is ascending)
    if sort_by_key:
        data = sorted(data,
                      key=itemgetter(sort_by_key),
                      reverse=sort_order_reverse)

    # If header is not empty, add header to data
    if header:
        # Get the length of each header and create a divider based
        # on that length
        header_divider = []
        for name in header:
            header_divider.append('-' * len(name))

        # Create a list of dictionary from the keys and the header and
        # insert it at the beginning of the list. Do the same for the
        # divider and insert below the header.
        header_divider = dict(zip(keys, header_divider))
        data.insert(0, header_divider)
        header = dict(zip(keys, header))
        data.insert(0, header)

    column_widths = []
    for key in keys:
        column_widths.append(max(len(str(column[key])) for column in data))

    # Create a tuple pair of key and the associated column width for it
    key_width_pair = zip(keys, column_widths)

    format = ('%-*s ' * len(keys)).strip() + '\n'
    formatted_data = ''
    for element in data:
        data_to_format = []
        # Create a tuple that will be used for the formatting in
        # width, value format
        for pair in key_width_pair:
            data_to_format.append(pair[1])
            data_to_format.append(element[pair[0]])
        formatted_data += format % tuple(data_to_format)
    return formatted_data


def charm_review_queue():
    print "Connecting to launchpad..."
    lp = Launchpad.login_anonymously('charm-tools', 'production',
                                     version='devel',
                                     launchpadlib_dir='~/.cache/launchpadlib')
    charm = lp.distributions['charms']
    charmers = lp.people['charmers']
    charm_contributors = lp.people['charm-contributors']

    print "Querying launchpad for bugs ..."
    bugs = charm.searchTasks(tags=['new-formula', 'new-charm'],
                             status=['New', 'Confirmed', 'Triaged',
                                     'In Progress', 'Fix Committed'],
                             tags_combinator="Any")
    charmers_bugs = charmers.searchTasks(
        status=['New', 'Confirmed', 'Triaged', 'In Progress',
                'Fix Committed'])

    print "Querying launchpad for proposals ..."
    proposals = charmers.getRequestedReviews(status="Needs review")
    charm_contributors_proposals = charm_contributors.getRequestedReviews(
        status="Needs review")

    print "Building review_queue ..."
    queue = list()
    max_summary_length = 50

    # Bugs in charms distribution and charmers group
    for bug in itertools.chain(bugs.entries, charmers_bugs.entries):
        entry_summary = bug['title'].split('"')[1].strip()
        bug_created = datetime.datetime.strptime(
            bug['date_created'].split('+')[0], "%Y-%m-%dT%H:%M:%S.%f")
        entry_age = datetime.datetime.utcnow() - bug_created
        entry = {'date_created': bug['date_created'].split("T")[0],
                 'age': str(entry_age).split('.')[0],
                 'summary': (entry_summary[:max_summary_length] + '...')
                 if len(entry_summary) > max_summary_length else entry_summary,
                 'item': bug['web_link'],
                 'status': bug['status'],
                 }
        queue.append(entry)

    # Merge proposals in charmers group
    for proposal in itertools.chain(proposals.entries,
                                    charm_contributors_proposals.entries):
        proposal_summary = proposal['description']
        proposal_date_created = datetime.datetime.strptime(
            proposal['date_created'].split('+')[0], "%Y-%m-%dT%H:%M:%S.%f")
        proposal_age = datetime.datetime.utcnow() - proposal_date_created
        if proposal_summary is None:
            proposal_summary = "Proposal"
        entry = {'date_created': proposal['date_created'].split("T")[0],
                 'age': str(proposal_age).split('.')[0],
                 'summary': (proposal_summary[:max_summary_length] + '...')
                 if len(proposal_summary) > max_summary_length
                 else proposal_summary,
                 'item': proposal['web_link'],
                 'status': proposal['queue_status'],
                 }
        queue.append(entry)

    return(sorted(queue, key=lambda k: k['date_created']))


def main():
    parser = argparse.ArgumentParser(
        description="Shows items needing the attention of ~charmers")
    parser.parse_args()
    review_queue = charm_review_queue()
    keys = ['date_created', 'age', 'summary', 'item', 'status']
    headers = ['Date Created', 'Age', 'Summary', 'Item', 'Status']
    print "Queue length: %d" % len(review_queue)
    if isinstance(review_queue, list) and len(review_queue) > 0:
        print format_as_table(review_queue,
                              keys,
                              header=headers,
                              sort_by_key='date_created',
                              sort_order_reverse=False)

if __name__ == "__main__":
    main()



#!/usr/bin/python
#
#    promulgate - makes a charm recipe branch the official one
#
#    Copyright (C) 2011  Canonical Ltd.
#    Author: Francis J. Lacoste <francis.lacoste@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from launchpadlib.launchpad import Launchpad
from lazr.restfulclient.errors import BadRequest, NotFound

import os
import sys

from optparse import OptionParser

from bzrlib import bzrdir

import yaml

import logging

import subprocess

DISTRIBUTION = 'charms'
REVIEW_TEAM_NAME = 'charmers'
OFFICIAL_BRANCH_POCKET = 'Release'
OFFICIAL_BRANCH_STATUS = 'Mature'


def parse_options():
    parser = OptionParser(usage='usage: %prog [options] <charm_dir>')

    parser.add_option(
        '-b', '--branch', dest='branch', default=None,
        help='The location of the charm public branch. Will be determined '
        'from the bzr configuration if omitted.')

    parser.add_option(
        '-s', '--series', dest='series', default=None,
        help='The distribution series on which to set the official branch. '
        'Defaults to setting it in the current development series.')

    parser.add_option(
        '-t', '--lp-instance', dest='lp_instance', default='production',
        help="The Launchpad instance to use. Defaults to production, but "
        "staging' or 'qastaging' might be used for testing.")

    parser.add_option(
        '-v', '--verbose', dest='verbose', action='count', default=0,
        help='Increase verbosity level.')

    parser.add_option(
        '-u', '--unpromulgate', dest='unpromulgate', action='store_true',
        default=False,
        help='Un-promulgate this branch instead of promulgating it')

    parser.add_option(
        '-f', '--force', dest='force', action='store_true', default=False,
        help='Override warnings and errors.  USE WITH EXTREME CARE !!!!')

    parser.add_option(
        '-w', '--ignore-warnings', dest='ignore_warnings', action='store_true',
        default=False,
        help='Promulgate this branch even with warnings from charm proof')

    parser.add_option(
        '-o', '--owner-branch', dest='promulgate_owner_branch',
        action='store_true', default=False,
        help='Promulgate a branch owned by a someone/group other than '
        '~charmers')

    return parser.parse_args()


def log_level(verbose):
    if verbose >= 2:
        return logging.DEBUG
    elif verbose >= 1:
        return logging.INFO
    else:
        return logging.WARNING


def proof_charm(charm_dir, force=False, ignore_warnings=False):
    logging.info("Running charm proof ... ")
    charm_proof = subprocess.call([sys.executable, os.path.join(
                                   os.path.dirname(os.path.realpath(__file__)),
                                   'proof.py'), charm_dir])
    if charm_proof == 1 or charm_proof == 200:
        if force:
            logging.info("force option enabled ... Continuing with errors")
        else:
            sys.exit(1)
    if charm_proof == 100:
        if ignore_warnings:
            logging.info("ignore-warnings enabled ... Continuing with "
                         "warnings")
        elif force:
            logging.info("force option enabled ... Continuing with warnings")
        else:
            sys.exit(charm_proof)
    if charm_proof == 0:
        logging.info("Excellent ... charm proof passed with flying colors")


def charm_name_from_metadata(charm_dir):
    charm_metadata = os.path.join(charm_dir, 'metadata.yaml')
    if not os.access(charm_metadata, os.R_OK):
        logging.error("can't read charm metadata: %s", charm_metadata)

    with open(charm_metadata) as metadata:
        charm = yaml.safe_load(metadata)

    return charm['name']


def find_branch_to_promulgate(lp, charm_dir, branch_url):
    if branch_url is None:
        tree, branch, relpath = bzrdir.BzrDir.open_containing_tree_or_branch(
            charm_dir)
        push_location = branch.get_push_location()
        if push_location is None:
            logging.error("Branch has not been pushed.")
            return 1
        charm_branch = lp.branches.getByUrl(url=push_location)
        if charm_branch is None:
            logging.error("can't determine Launchpad branch from bzr branch")
            return 1
    else:
        charm_branch = lp.branches.getByUrl(url=branch_url)
        if charm_branch is None:
            logging.error("can't find branch on Launchpad: %s", branch_url)
            return 1
    return charm_branch


def get_lp_charm_series(lp, series):
    charm_distro = lp.distributions[DISTRIBUTION]
    if series is None:
        charm_series = charm_distro.current_series
    else:
        try:
            charm_series = charm_distro.getSeries(
                name_or_version=series)
        except (BadRequest, NotFound), e:
            # XXX flacoste 2011-06-15 bug=797917
            # Should only be NotFound.
            if e.content.startswith('No such distribution series:'):
                logging.error("can't find series '%s'", series)
                raise
            else:
                raise
    return charm_series


def update_branch_info(charm_branch, branch_status, branch_reviewer):
    logging.info("Setting status of %s to %s", charm_branch.bzr_identity,
                 branch_status)
    charm_branch.lifecycle_status = branch_status

    logging.info("Setting reviewer of %s to %s", charm_branch.bzr_identity,
                 branch_reviewer)
    charm_branch.reviewer = branch_reviewer

    charm_branch.lp_save()


def update_official_charm_branch(lp, series, charm_branch, charm_name):
    charm_series = get_lp_charm_series(lp, series)
    lp_charm = charm_series.getSourcePackage(name=charm_name)
    if charm_branch:
        logging.info('Setting %s as the official branch for %s',
                     charm_branch.bzr_identity,
                     lp_charm.name)
        update_branch_info(charm_branch, OFFICIAL_BRANCH_STATUS,
                           lp.people[REVIEW_TEAM_NAME])
    else:
        logging.info('Removing official branch for %s', lp_charm.name)
    lp_charm.setBranch(branch=charm_branch, pocket=OFFICIAL_BRANCH_POCKET)


def branch_owner(bzr_branch):
    lp_url = bzr_branch.bzr_identity  # TODO this really sucks... better way?
    return lp_url.lstrip('lp:').split('/')[0]


def is_valid_owner(charm_branch, promulgate_owner_branch):
    if charm_branch is None:
        return True
    return promulgate_owner_branch or branch_owner(charm_branch) == '~charmers'


def main():
    options, args = parse_options()
    logging.basicConfig(level=log_level(options.verbose),
                        format='%(levelname)s:%(message)s')

    if len(args):
        charm_dir = args[0]
    else:
        charm_dir = os.getcwd()

    proof_charm(charm_dir, options.force, options.ignore_warnings)

    logging.debug('login with %s launchpad:', options.lp_instance)
    lp = Launchpad.login_with('promulgate', options.lp_instance)

    if options.unpromulgate:
        logging.info('unpromulgating...')
        charm_branch = None  # makes LP delete the source package.
    else:
        logging.info('promulgating...')
        charm_branch = find_branch_to_promulgate(lp, charm_dir, options.branch)

    if not is_valid_owner(charm_branch, options.promulgate_owner_branch):
        logging.error(" Invalid branch owner: %s", branch_owner(charm_branch))
        logging.error(" Branch push location must be owned by '~charmers'\n"
                      "   use `bzr push --remember lp:~charmers/charms/"
                      "<series>/<charm-name>/trunk`\n   or override this "
                      "behavior using the '--owner-branch'"
                      " option")
        return 1

    update_official_charm_branch(lp, options.series, charm_branch,
                                 charm_name_from_metadata(charm_dir))

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
#
#    create - generate Juju charm from template
#
#    Copyright (C) 2011  Canonical Ltd.
#    Author: Clint Byrum <clint.byrum@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import os.path as path
import time
import shutil
import tempfile
import textwrap
import socket
import argparse

from Cheetah.Template import Template
from stat import ST_MODE


def portable_get_maintainer():
    """ Portable best effort to determine a maintainer """
    if 'NAME' in os.environ:
        name = os.environ['NAME']
    else:
        try:
            import pwd
            name = pwd.getpwuid(os.getuid()).pw_gecos.split(',')[0].strip()

            if not len(name):
                name = pwd.getpwuid(os.getuid())[0]
        except:
            name = 'Your Name'

    if not len(name):
        name = 'Your Name'

    email = os.environ.get('EMAIL', '%s@%s' % (name.replace(' ', '.'),
                                               socket.getfqdn()))
    return name, email


def setup_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('charmname', help='Name of charm to create.')
    parser.add_argument('charmhome', nargs='?',
                        help='Dir to create charm in. Defaults to CHARM_HOME '
                        'env var or PWD')

    return parser


def apt_fill(package):
    v = {}
    try:
        import apt
        c = apt.Cache()
        c.open()
        p = c[package]
        print "Found " + package + " package in apt cache, as a result charm" \
              + " contents have been pre-populated based on package metadata."

        v['summary'] = p.summary
        v['description'] = textwrap.fill(p.description, width=72,
                                         subsequent_indent='  ')
    except:
        print "Failed to find " + package + " in apt cache, creating " \
            + "an empty charm instead."
        v['summary'] = '<Fill in summary here>'
        v['description'] = '<Multi-line description here>'

    return v


def main():
    parser = setup_parser()
    args = parser.parse_args()

    try:
        from ubuntutools.config import ubu_email as get_maintainer
    except ImportError:
        get_maintainer = portable_get_maintainer

    if args.charmhome:
        charm_home = args.charmhome
    else:
        charm_home = os.getenv('CHARM_HOME', '.')

    home = path.abspath(path.dirname(__file__))
    template_dir = path.join(home, 'templates')
    output_dir = path.join(charm_home, args.charmname)
    print "Generating template for " + args.charmname + " from templates in " \
        + template_dir
    print "Charm will be stored in " + output_dir

    if path.exists(output_dir):
        print output_dir + " exists. Please move it out of the way."
        sys.exit(1)

    shutil.copytree(path.join(template_dir, 'charm'), output_dir)

    v = {'package': args.charmname,
         'maintainer': '%s <%s>' % get_maintainer()}

    v.update(apt_fill(args.charmname))

    ignore_parsing = ['README.ex']

    for root, dirs, files in os.walk(output_dir):
        for outfile in files:
            full_outfile = path.join(root, outfile)
            mode = os.stat(full_outfile)[ST_MODE]
            if outfile in ignore_parsing:
                continue

            try:
                t = Template(file=full_outfile, searchList=(v))
                o = tempfile.NamedTemporaryFile(dir=root, delete=False)
                os.chmod(o.name, mode)
                o.write(str(t))
                o.close()
                backupname = full_outfile + str(time.time())
                os.rename(full_outfile, backupname)
                try:
                    os.rename(o.name, full_outfile)
                    os.unlink(backupname)
                except Exception, e:
                    print "WARNING: Could not enable templated file: " + str(e)
                    os.rename(backupname, full_outfile)
                    raise
            except Exception, e:
                print "WARNING: could not process template for " \
                    + full_outfile + ": " + str(e)
                raise

if __name__ == "__main__":
    main()



#!/usr/bin/python
#
#    subscribers - list and manipulate charm subscribers
#
#    Copyright (C) 2012  Canonical Ltd.
#    Author: Clint Byrum <clint.byrum@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import yaml
import argparse
import logging
from logging import warn, info, debug

from launchpadlib import launchpad


def setup_parser():
    parser = argparse.ArgumentParser(
        description="""Prints out a report and optionally corrects found
        instances where maintainers of charms are not subscribed to bugs
        on their charm in launchpad.

        Users will need all charms they are interested in checking locally
        in the specified repository so that metadata.yaml can be inspected
        to find the maintainer.""")
    parser.add_argument('--subscribed', default=False, action='store_true',
                        help='Show maintainers who are properly subscribed '
                        'instead of unsubscribed')
    parser.add_argument('--repository', default=None, type=str,
                        help='Repository to look for charms in. Defaults to '
                        '$JUJU_REPOSITORY or getcwd')
    parser.add_argument('--quiet', default=False, action='store_true',
                        help='Hide everything except maintainer subscription '
                        'lists.')
    parser.add_argument('--series', '-s', default=None,
                        help='Which series of the charm store to run against. '
                        'Defaults to current dev focus')
    parser.add_argument('--maintainer', default=None,
                        help='Limit output to this maintainer\'s charms only.')
    parser.add_argument('--log-priority', default='WARNING')
    parser.add_argument('--launchpad-instance', default='production')
    parser.add_argument('--fix-unsubscribed', default=False,
                        action='store_true', help='Add a bug subscription for '
                        'any unsubscribed maintainers. Requires --maintainer')
    parser.add_argument('--force-fix-all', default=False, action='store_true',
                        help=argparse.SUPPRESS)
    parser.add_argument('charms', default=[], nargs='*',
                        help='Charms to check for subscriptions')

    return parser


def main():
    parser = setup_parser()
    args = parser.parse_args()

    if args.repository is None:
        repository = os.environ.get('JUJU_REPOSITORY', os.getcwd())
    else:
        repository = args.repository

    if args.log_priority == 'WARNING':
        log_prio = logging.WARNING
    elif args.log_priority == 'INFO':
        log_prio = logging.INFO
    elif args.log_priority == 'CRITICAL':
        log_prio = logging.CRITICAL
    elif args.log_priority == 'INFO':
        log_prio = logging.INFO
    elif args.log_priority == 'DEBUG':
        log_prio = logging.DEBUG
    else:
        log_prio = logging.DEBUG

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                        level=log_prio)

    if args.quiet:
        logging.disable(logging.WARNING)

    if args.maintainer is None and args.fix_unsubscribed:
        warn('Running --fix-unsubscribed and without --maintainer is against'
             ' policy.')
        if not args.force_fix_all:
            warn('Use --force-fix-all to override policy')
            sys.exit(1)
        else:
            warn('--force-fix-all passed, continuing')

    app_name = 'Charm Tools - subscribers'
    home_dir = os.environ.get('HOME', os.path.expanduser('~'))
    launchpadlib_dir = os.path.join(home_dir, '.cache', 'launchpadlib')

    if args.fix_unsubscribed:
        login = launchpad.Launchpad.login_with
    else:
        login = launchpad.Launchpad.login_anonymously

    lp = login(app_name, args.launchpad_instance,
               version='devel', launchpadlib_dir=launchpadlib_dir)

    charmdistro = lp.distributions['charms']

    if args.series:
        current_series = args.series
    else:
        current_series = str(charmdistro.current_series).split('/').pop()

    charms = []
    if len(args.charms):
        for charm_name in args.charms:
            # XXX Is this line really neccesary?
            charmdistro.getSourcePackage(name=charm_name)
            charms.append(charm_name)
    else:
        branches = charmdistro.getBranchTips()
        for branch in branches:
            try:
                branch_series = str(branch[2][0]).split('/')[0]
                charm_name = str(branch[0]).split('/')[3]
            except IndexError:
                continue
            if branch_series != current_series:
                continue
            charms.append(charm_name)

    for charm_name in charms:
        try:
            with open('%s/%s/%s/metadata.yaml' % (repository, current_series,
                                                  charm_name)) as mdata:
                mdata_parsed = yaml.safe_load(mdata)
        except IOError:
            warn('%s/%s has no metadata in charm repo' % (current_series,
                                                          charm_name))
            continue
        try:
            maintainers = mdata_parsed['maintainer']
        except KeyError:
            warn('%s has no maintainer' % charm_name)
            continue

        if type(maintainers) == str:
            maintainers = [maintainers]

        if args.maintainer is not None:
            if args.maintainer not in maintainers:
                maints_by_email = [m.split('<')[1].split('>')[0]
                                   for m in maintainers]
                if args.maintainer not in maints_by_email:
                    debug('%s not in maintainer list %s' % (args.maintainer,
                                                            maintainers))
                    continue

        for maintainer in maintainers:
            maint_email = maintainer.split('<')[1].split('>')[0]
            lp_maintainer = lp.people.getByEmail(email=maint_email)
            if not lp_maintainer:
                warn('%s has no people in launchpad' % maintainer)
                continue
            pkg = charmdistro.getSourcePackage(name=charm_name)
            subscription = pkg.getSubscription(person=lp_maintainer)
            if subscription is not None:
                msg = '%s is subscribed to %s' % (maintainer, charm_name)
                if args.subscribed:
                    print msg
                else:
                    info(msg)
            elif not args.subscribed:
                msg = '%s is not subscribed to %s' % (maintainer, charm_name)
                if args.subscribed:
                    info(msg)
                else:
                    print msg
                if args.fix_unsubscribed:
                    info('adding bug subscription to %s for %s' %
                         (charm_name, maint_email))
                    pkg.addBugSubscription(subscriber=lp_maintainer)



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess


def main():
    exit_code = subprocess.call([sys.executable, os.path.join(os.path.dirname(
                                 os.path.realpath(__file__)), 'promulgate'),
                                '--unpromulgate'] + sys.argv[1:])
    sys.exit(exit_code)



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

import subprocess

from . import cli
from . import version


def charm():
    if len(sys.argv) < 2:
        cli.usage(1)

    sub = sys.argv[1]
    opts = sys.argv[2:]
    if sub == '--description':
        sys.stdout.write("Tools for authoring and maintaining charms\n")
        sys.exit(0)

    if sub == '--help':
        cli.usage(0)

    if sub == '--version':
        version.main()
        sys.exit(0)

    if sub == '--list':
        print '\n'.join(cli.subcommands(os.path.realpath(__file__)))
        sys.exit(0)

    sub_exec = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])),
                            "charm-%s%s" % (sub, cli.ext()))

    if not os.path.exists(sub_exec):
        sys.stderr.write('Error: %s is not a valid subcommand\n\n' % sub)
        cli.usage(2)
    subprocess.call([sub_exec] + opts)


def bundle():
    if len(sys.argv) < 2:
        cli.usage(0)

    sub = sys.argv[1]
    opts = sys.argv[2:]

    if sub == '--description':
        sys.stdout.write("Tools for managing bundles\n")
        sys.exit(0)

    if sub == '--help':
        cli.usage(0)

    sub_exec = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])),
                            "charm-%s%s" % (sub, cli.ext()))

    if not os.path.exists(sub_exec):
        sys.stderr.write('Error: %s is not a valid subcommand\n\n' % sub)
        cli.usage(2)

    subprocess.call([sub_exec, '--bundle'] + opts)


if __name__ == '__main__':
    charm()



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import argparse

from mr import Mr
from charmworldlib import charm as cwc


def setup_parser():
    parser = argparse.ArgumentParser(prog='charm get',
                                     description='Retrieves official charm '
                                                 'branch from launchpad.net')
    parser.add_argument('charm', nargs=1, help='Charm to branch',
                        metavar=('charm_name'))
    parser.add_argument('branch_to', nargs='?',
                        help='Path to where charm should be branched')

    return parser


def parse_charm_id(charm_id):
    if charm_id.startswith('cs:'):
        charm_id = charm_id.replace('cs:', '')

    try:
        charm = cwc.Charm(charm_id)
    except:
        return None

    return charm


def download(charm, to):
    charm_dir = os.path.join(to, charm.name)
    if os.path.exists(charm_dir) and os.listdir(charm_dir):
        raise ValueError('%s exists and is not empty' % charm_dir)

    if not os.path.exists(to):
        os.makedirs(to)

    mr = Mr(to, mr_compat=False)
    mr.add(charm.name, charm.code_source['location'], checkout=True)


def main(args=None):
    parser = setup_parser()
    args = parser.parse_args(args)

    charm = parse_charm_id(args.charm[0])

    if not charm:
        sys.stderr.write('Error: %s not found in charm store\n'
                         % args.charm[0])
        sys.exit(1)

    ldir = args.branch_to
    branch_dir = os.path.realpath(ldir) if ldir else os.getcwd()

    sys.stderr.write('Branching %s (%s) to %s/%s\n' % (charm.name,
                     charm.code_source['location'], branch_dir,
                     charm.name))

    try:
        download(charm, branch_dir)
    except Exception as e:
        sys.stderr.write('Error: %s' % str(e))
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/python
# coding=latin-1

import argparse
import glob
import logging
import os
import re
import signal
import subprocess
import sys
import time
import yaml
from datetime import timedelta
from contextlib import contextmanager

TEST_PASS = '✔'
TEST_FAIL = '✘'
TEST_SKIP = '↷'
TEST_TIMEOUT = '⌛'
TEST_STATUS = {'pass': TEST_PASS, 'fail': TEST_FAIL, 'skip': TEST_SKIP,
               'timeout': TEST_TIMEOUT}
TEST_RESERVED_EXITS = {0: 'pass', 100: 'skip', 124: 'timeout'}

LOG_LEVELS = [logging.INFO, logging.DEBUG]
TEST_RESULT_LEVELV_NUM = 51


class NoTests(Exception):
    pass


class BootstrapError(Exception):
    pass


class BootstrapUnreliable(Exception):
    pass


class DestroyUnreliable(Exception):
    pass


class SubstrateMismatch(Exception):
    pass


class TimeoutError(Exception):
    def __init__(self, value="Timed Out"):
        self.value = value


class TestingError(Exception):
    pass


class OrchestraError(Exception):
    pass


class Conductor(object):
    def __init__(self, arguments=None):
        self.args = arguments
        self.env = {'JUJU_HOME': os.path.expanduser('~/.juju')}
        self.log = logging.getLogger('juju-test.conductor')
        self.tests = self.args.tests or self.find_tests()
        self.tests_supplied = bool(self.args.tests)
        self.juju_version = None
        self.juju_env = self.args.juju_env
        self.errors = 0
        self.fails = 0
        self.passes = 0

        if not self.tests:
            raise NoTests()

    def run(self):
        self.juju_version = get_juju_version()
        available_tests = self.find_tests()

        for test in self.tests:
            if not os.path.basename(test) in available_tests:
                self.log.error('%s is not an available test. Skipping' % test)
                self.errors += 1
                continue

            if not os.path.isfile(test):
                if not os.path.isfile(os.path.join('tests', test)):
                    self.log.error('%s was not found. Skipping' % test)
                    self.errors += 1
                    continue
                else:
                    test = os.path.join('tests', test)

            try:
                self.bootstrap(self.juju_env, self.args.setup_timeout)
            except Exception, e:
                self.log.warn('Could not bootstrap %s, got %s. Skipping' %
                              (self.juju_env, e))
                self.errors += 1
                continue

            try:
                t = Orchestra(self, test)
                t.perform()
            except:
                self.fails += 1
                if self.args.set_e:
                    self.log.info('Breaking here as requested by --set-e')
                    return self.errors, self.fails, self.passes
            else:
                self.passes += 1

            try:
                self.destroy(self.juju_env)
            except DestroyUnreliable:
                self.log.warn('Unable to destroy bootstrap, trying again')
                time.sleep(2)
                try:
                    self.destroy(self.juju_env)
                except:
                    continue

        return self.errors, self.fails, self.passes

    def find_tests(self):
        tests_dir = glob.glob(os.path.join('tests', '*'))
        if not tests_dir:
            return None

        # Filter out only the files in tests/ then get the test names.
        tests = [os.path.basename(t) for t in tests_dir if os.path.isfile(t)]

        return sorted(tests)

    def safe_test_name(self, test_name):
        return test_name

    def isolate_environment(self, juju_env):
        # Should probably do something other than NotImplementedError...
        raise NotImplementedError()

    def get_environment(self, juju_env, juju_home='~/.juju'):
        try:
            env_yaml = self.load_environments_yaml(juju_home)
        except IOError:
            raise  # Do something more clever here?

        if not juju_env in env_yaml['environments']:
            raise KeyError('%s does not exist in %s/environments.yaml' %
                           (juju_env, juju_home))

        return env_yaml['environments'][juju_env]

    def bootstrap(self, juju_env=None, wait_for=400):
        self.log.debug('Starting a bootstrap for %s, kill after %s'
                       % (juju_env, wait_for))
        cmd = ['juju', 'bootstrap']
        if self.juju_version.major > 0:
            if self.args.upload_tools:
                cmd.append('--upload-tools')
        if self.args.constraints:
            cmd.extend(['--constraints', self.args.constraints])
        cmd.extend(['-e', juju_env])

        self.log.debug('Running the following: %s' % ' '.join(cmd))
        try:
            subprocess.check_output(cmd)
        except subprocess.CalledProcessError:
            raise BootstrapError('Bootstrap returned with an exit > 0')

        self.log.debug('Waiting for bootstrap')
        try:
            with timeout(wait_for):
                self.wait_for_bootstrap(juju_env)
        except TimeoutError:
            try:
                self.destroy(self.juju_env)
            except:
                pass

            raise BootstrapUnreliable('Bootstrap timeout after %ss' % wait_for)

    def destroy(self, juju_env):
        self.log.debug('Tearing down %s juju environment' % juju_env)
        cmd = ['juju', 'destroy-environment']

        if self.juju_version.major > 0:
            if self.juju_version.minor < 17:
                cmd.extend(['-y', '-e', juju_env])
            else:
                cmd.extend(['-y', juju_env])

            self.log.debug('Calling "%s"' % ' '.join(cmd))
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError:
                raise DestroyUnreliable('Unable to destroy %s' % juju_env)
        else:
            # Probably should use Popen instead of Shell=True. I'm just not
            # confident on properly mocking a Popen call just yet.
            cmd.extend(['-e', juju_env])
            pycmd = 'echo y | %s' % ' '.join(cmd)
            self.log.debug('Calling "%s"' % pycmd)
            try:
                subprocess.check_call(pycmd, shell=True)
            except subprocess.CalledProcessError:
                raise DestroyUnreliable('Unable to destroy %s' % juju_env)

    def status(self, juju_env):
        cmd = ['juju', 'status', '-e', juju_env]
        self.log.debug('Running the following: %s' % ' '.join(cmd))
        try:
            output = subprocess.check_output(cmd)
        except:
            self.log.debug('Status command failed, returning nothing')
            return None

        return yaml.safe_load(output)

    def wait_for_bootstrap(self, juju_env=None):
        bootstrapped = False
        while not bootstrapped:
            self.log.debug('Still not bootstrapped')
            time.sleep(5)
            status = self.status(juju_env)
            if not status:
                continue

            if self.juju_version.major > 0:
                self.log.debug('State for %s: %s' %
                               (self.juju_version,
                                status['machines']['0']['agent-state']))
                if status['machines']['0']['agent-state'] == 'started':
                    bootstrapped = True
            else:
                self.log.debug('State for %s: %s' %
                               (self.juju_version,
                                status['machines'][0]['agent-state']))
                if status['machines'][0]['agent-state'] == 'running':
                    bootstrapped = True

    def load_environments_yaml(self, juju_home='~/.juju'):
        env_yaml_file = os.path.join(os.path.expanduser(juju_home),
                                     'environments.yaml')
        if not os.path.exists(env_yaml_file):
            raise IOError("%s file does not exist" % env_yaml_file)
        with open(env_yaml_file, 'r') as y:
            return yaml.safe_load(y.read())


class Orchestra(object):
    def __init__(self, conductor, arrangement):
        self.conductor = conductor
        self.test = arrangement
        self.name = os.path.basename(self.test)
        self.safe_name = self.conductor.safe_test_name(self.name)
        self.log = logging.getLogger('juju-test.conductor.%s' % self.safe_name)
        self.archive = self.conductor.args.logdir
        self.env = self.conductor.env

    def perform(self):
        self.build_env()
        error = None
        self.log.debug('Running %s (%s)' % (self.name, self.test))
        try:
            with timeout(self.conductor.args.timeout):
                output = subprocess.check_output(self.test, env=self.env)
                self.log.debug(output)
        except TimeoutError, e:
            self.log.debug('Killed by timeout after %s seconds',
                           self.conductor.args.timeout)
            self.print_status(124)
            error = e if not self.is_passing_code(124) else e
        except subprocess.CalledProcessError, e:
            self.log.debug('Got exit code: %s', e.returncode)
            self.print_status(e.returncode)
            error = TestingError(e.returncode) if not \
                self.is_passing_code(e.returncode) else e
        except Exception, e:
            self.log.debug('Encountered unexpected error %s', e)
            self.print_status(9001)
            error = e
        else:
            self.print_status(0)

        if self.conductor.args.logdir:
            try:
                self.archive_logs()
            except OrchestraError, e:
                self.log.error(e)

        if error:
            raise error

    def archive_logs(self):
        logs = ['/var/./log/juju/*']
        status = self.conductor.status(self.env["JUJU_ENV"])
        logdir = self.conductor.args.logdir
        if not status:
            # Something is wrong, we need to throw up an archive error
            raise OrchestraError('Unable to query juju status')

        services = status['services']
        # machines = status['machines']

        if self.conductor.juju_version.major == 0:
            logs.append('/var/lib/juju/units/./*/charm.log')

        try:
            self.rsync(0, logs[0], os.path.join(logdir, 'bootstrap', ''))
        except:
            self.log.warn('Failed to fetch logs for bootstrap node')

        for service in services:
            for unit in services[service]['units']:
                machine = services[service]['units'][unit]['machine']
                for log in logs:
                    try:
                        self.rsync(machine, log, os.path.join(logdir, service,
                                                              ''))
                    except:
                        self.log.warn('Failed to grab logs for %s' % unit)

    def print_status(self, exit_code):
        actual = self.map_status_code(exit_code)
        computed = self.determine_status(exit_code)

        if actual != computed:
            self.log.status('%s (%s)' % (TEST_STATUS[computed],
                                         TEST_STATUS[actual]))
        else:
            self.log.status(TEST_STATUS[actual])

    # The next three methods need to be collapsed and consolidated
    def determine_status(self, exit_code):
        if exit_code == 124:
            timeout_status = self.conductor.args.on_timeout
            reversed_codes = {v: k for k, v in TEST_RESERVED_EXITS.items()}
            if timeout_status in reversed_codes.keys():
                exit_code = reversed_codes[self.conductor.args.on_timeout]
            else:
                return timeout_status

        if not exit_code in TEST_RESERVED_EXITS.keys():
            return 'fail'

        if exit_code == 100 and self.conductor.args.fail_on_skip:
            return 'fail'
        elif exit_code == 100:
            return 'skip'

        if exit_code == 0:
            return 'pass'

    def is_passing_code(self, exit_code):
        if exit_code == 124:
            timeout_status = self.conductor.args.on_timeout
            reversed_codes = {v: k for k, v in TEST_RESERVED_EXITS.items()}
            if timeout_status in reversed_codes.keys():
                exit_code = reversed_codes[self.conductor.args.on_timeout]

        if exit_code == 0:
            return True
        if exit_code == 100 and not self.conductor.args.fail_on_skip:
            return True

        return False

    def map_status_code(self, exit_code):
        if exit_code in TEST_RESERVED_EXITS.keys():
            return TEST_RESERVED_EXITS[exit_code]
        else:
            return 'fail'

    def build_env(self):
        self.env["JUJU_ENV"] = self.conductor.juju_env

    def rsync(self, machine, path, dest):
        if self.conductor.juju_version.major == 0:
            cmd = ['rsync', '-a', '-v', '-z', '-R', '-e', 'juju ssh -e %s' %
                   self.env['JUJU_ENV'], '%s:%s' % (machine, path), dest]
        else:  # http://pad.lv/1183159
            status = self.conductor.status(self.env['JUJU_ENV'])
            dns_name = status['machines'][machine]['dns-name']
            cmd = ['rsync', '-a', '-v', '-z', '-R', '-e', 'ssh',
                   'ubuntu@%s:%s' % (dns_name, path), dest]

        subprocess.check_call(cmd)


# Build Juju class instead? Move bootstrap, wait_for_bootstrap, teardown?
class JujuVersion(object):
    def __init__(self, major=0, minor=0, patch=0):
        self.mapping = ['major', 'minor', 'patch']
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self):
        return '.'.join(str(v) for v in [self.major, self.minor, self.patch])


class TestCfg(object):
    _keys = ['timeout', 'set-e', 'on-timeout', 'fail-on-skip', 'tests']

    def __init__(self, cfg):
        if isinstance(cfg, basestring):
            cfg = yaml.safe_load(cfg)

        if 'options' in cfg:
            for key, val in cfg['options'].iteritems():
                if key in self._keys:
                    setattr(self, key, val)
        if 'substrates' in cfg:
            self.substrates = cfg.substrates


def get_juju_version():
    jv = JujuVersion()
    cmd = ['juju', 'version']
    try:
        version = subprocess.check_output(cmd)
        version = version.split('-')[0]
    except:
        cmd[1] = '--version'
        version = subprocess.check_output(cmd)
        version = version.split()[1]

    for i, ver in enumerate(version.split('.')):
        try:
            setattr(jv, jv.mapping[i], int(ver))
        except:
            break  # List out of range? Versions not semantic? Not my problem.

    return jv


def setup_logging(level=0, quiet=False, logdir=None):
    logger = logging.getLogger('juju-test')
    ft = logging.Formatter('%(asctime)s %(name)s %(levelname)-8s: %(message)s')
    cmt = logging.Formatter('%(name)s %(levelname)-8s: %(message)s')
    logger.setLevel(1)

    if level >= len(LOG_LEVELS):
        level = len(LOG_LEVELS) - 1

    if logdir:
        if not os.path.exists(logdir):
            os.makedirs(logdir)
        logfile = os.path.join(logdir, 'juju-test.%s.log' % int(time.time()))
        fh = logging.FileHandler(logfile)
        # Always at least log to INFO for file, unless DEBUG is requested
        fh.setLevel(LOG_LEVELS[level])
        fh.setFormatter(ft)
        logger.addHandler(fh)

    if not quiet:
        ch = logging.StreamHandler()
        ch.setLevel(LOG_LEVELS[level])
        ch.setFormatter(cmt)
        logger.addHandler(ch)

    return logger


class SubstrateFilter(object):
    def __init__(self, spec):
        self.order = spec.get('order', ['include', 'skip'])
        self.include = spec.get('include', ['*'])
        self.skip = spec.get('skip', [])

        if isinstance(self.order, str):
            self.order = [s.strip() for s in self.order.split(',')]
        if self.order != ['include', 'skip'] and \
                self.order != ['skip', 'include']:
            raise ValueError(
                'order should be defined using only include and skip')

        if isinstance(self.include, str):
            self.include = [self.include]
        self.include = set(self.include)

        if isinstance(self.skip, str):
            self.skip = [self.skip]
        self.skip = set(self.skip)

    def filter(self, substrates):
        """
        Filter a list of substrates relative to the rules generated on class
        creation.
        """
        if isinstance(substrates, str):
            substrates = [s.strip() for s in re.split('[,\s]', substrates)]

        # Apply the rules to the list of substrates returning anything that
        # matches
        if self.order == ['include', 'skip']:
            result = self._filter_includes(substrates, True)
            result = self._filter_skips(result)
        else:
            result = self._filter_skips(substrates, True)
            result = self._filter_includes(result)
        return result

    def _filter_includes(self, inputList, priority=False):
        if priority and '*' in self.include:
            return inputList
        return sorted(list(set(inputList).intersection(self.include)))

    def _filter_skips(self, inputList, priority=False):
        if priority and '*' in self.skip:
            return list(self.include.intersection(inputList))
        return sorted(list(set(inputList).difference(self.skip)))


def parse_substrates(spec):
    if isinstance(spec, basestring):
        spec = yaml.safe_load(spec)
    if not spec or 'substrates' not in spec:
        raise ValueError(
            "Invalid data passed to parse_substrates: {}".format(spec))

    specRules = SubstrateFilter(spec['substrates'])
    return specRules


def allowed_substrates(spec, possible_substrates):
    return parse_substrates(spec).filter(possible_substrates)


def setup_parser():
    parser = argparse.ArgumentParser(
        prog='juju test',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='Execute a charms functional tests',
        epilog="""\
`%(prog)s` should always be run from within a CHARM_DIR.

examples:

  Given the following example charm layout:
    .
    ├── config.yaml
    ├── copyright
    ├── hooks
    │   └── ...
    ├── icon.svg
    ├── metadata.yaml
    ├── README.md
    └── tests
        ├── 00-tool_setup
        ├── 01-standard
        ├── 02-at_scale
        └── 03-different_database

  Run all tests for current charm
    %(prog)s

  Run one or more tests
    %(prog)s 01-standard 03-different_database

output:

  Each unit test will return an output in the form or either:

    RESULT   : SYM

  or

    RESULT   : SYM (SYM)

  Where SYM is a Symbol representing PASS: ✔, FAIL: ✘, SKIP: ↷, or TIMEOUT: ⌛
  In the event a status is rewritten by either the --fail-on-skip flag or the
  --on-timeout flag the original status will be displayed in () next to the
  computed status.
  """)
    # Plugin specific
    parser.add_argument('--description', action="store_true",
                        help="produces one-line description for juju-core")
    # Tester specific
    # Options ommited from jitsu: --archive-only, --no-bootstrap
    parser.add_argument('--timeout', default=600, action=StoreTimedelta,
                        help="timeout per unit test. Examples: 10m, 300s")
    parser.add_argument('--isolate', metavar='JUJU_ENV',
                        help="create unique environment cloned from JUJU_ENV")
    parser.add_argument('-o', '--logdir',
                        help="directory to store service logs from each test")
    # New tester options
    parser.add_argument('--setup-timeout', default=300, action=StoreTimedelta,
                        help="timeout to wait for an environment to be set up")
    parser.add_argument('--fail-on-skip', default=False, action="store_true",
                        help="treat SKIP (100) status as a failure, will \
                              halt execution with --set-e")
    parser.add_argument('--on-timeout', default='skip',
                        choices=['fail', 'skip', 'pass'],
                        help="treat tests which timeout as (fail, skip, pass)")
    parser.add_argument('--set-e', default=False, action="store_true",
                        help="stop testing execution on first failed test")
    parser.add_argument('-v', action='count', default=0,
                        help="make test more verbose")
    parser.add_argument('-q', '--quiet', default=False, action="store_true",
                        help="quiet all output")
    # These are bootstrap/juju specific
    parser.add_argument('-e', '--environment', metavar='JUJU_ENV',
                        default=os.environ.get('JUJU_ENV'),
                        dest='juju_env',
                        #required=True,
                        help="juju environment to operate in")
    parser.add_argument('--upload-tools', default=False, action="store_true",
                        help="upload juju tools (available for juju > 1.x)")
    parser.add_argument('--constraints', help="juju environment constraints")
    # General options
    # TODO: Need to add some way to specify a particular test without
    #       colliding with CHARM
    parser.add_argument('tests', metavar="TEST", nargs='*',
                        help="tests to execute, relative to tests/ directory, \
                              default is all")
    return parser


# http://stackoverflow.com/a/11784984/196832
def status(self, message, *args, **kws):
    self._log(TEST_RESULT_LEVELV_NUM, message, args, **kws)

logging.addLevelName(TEST_RESULT_LEVELV_NUM, "RESULT")
logging.Logger.status = status


def main():
    p = setup_parser()
    args = p.parse_args()
    test_cfg = None
    if args.description:
        print p.description
        sys.exit()

    logger = setup_logging(level=args.v, quiet=args.quiet, logdir=args.logdir)
    logger.info('Starting test run on %s using Juju %s'
                % (args.juju_env, get_juju_version()))
    logger.debug('Loading configuration options from testplan YAML')
    test_plans = glob.glob(os.path.join(os.getcwd(), 'tests',
                                        'test_config.y*ml'))
    if test_plans:
        with open(test_plans[0]) as f:
            test_cfg = f.read()

    if test_cfg:
        cfg = TestCfg(test_cfg)
        for key, val in args.iteritems():
            logger.debug('Overwriting %s to %s from cmd' % (key, val))
            setattr(cfg, key, val)
    else:
        cfg = args

    logger.debug('Creating a new Conductor')
    try:
        tester = Conductor(args)
        env_yaml = tester.get_environment(cfg.juju_env)
        if 'substrates' in cfg:
            rules = parse_substrates(cfg)
            allowed = rules.filter(env_yaml['type'])
            if env_yaml['type'] not in allowed:
                raise Exception('%s is not an allowed substrate: %s' %
                                (env_yaml['type'],
                                 allowed.join(', ')))
        errors, failures, passes = tester.run()
    except NoTests:
        logger.critical('No tests were found')
        sys.exit(1)
    except Exception as e:
        logger.critical(str(e))
        sys.exit(1)
    except:
        raise

    logger.info('Results: %s passed, %s failed, %s errored' %
                (passes, failures, errors))
    if failures > 0:
        sys.exit(failures)
    elif errors > 0:
        sys.exit(124)  # Nothing failed, but there were errors!

    sys.exit(0)


# http://stackoverflow.com/a/601168/196832
@contextmanager
def timeout(seconds):
    def signal_handler(signum, frame):
        raise TimeoutError()
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class StoreTimedelta(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, convert_to_timedelta(values))


# http://bit.ly/13Mi5QV
def convert_to_timedelta(time_val):
    if isinstance(time_val, int):
        return time_val

    if time_val.isdigit():
        return int(time_val)

    num = int(time_val[:-1])
    if time_val.endswith('s'):
        return timedelta(seconds=num).seconds
    elif time_val.endswith('m'):
        return timedelta(minutes=num).seconds
    elif time_val.endswith('h'):
        return timedelta(hours=num).seconds


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import argparse

from cli import parser_defaults
from charmworldlib import charm


def info(charm_id, is_bundle=False):
    c = charm.Charm(charm_id)
    readme = [s for s in c.files if re.match("^readme", s, re.IGNORECASE)]

    if not readme:
        raise Exception('No README found')

    return c.file(readme[0])


def setup_parser(args=None):
    parser = argparse.ArgumentParser(
        description='Learn more about a Charm or Bundle')
    parser.add_argument('charm', nargs=1, help='Charm to branch',
                        metavar=('charm_name'))
    parser = parser_defaults(parser)

    return parser.parse_args(args)


def main():
    a = setup_parser()
    charm_id = a.charm[0].replace('cs:', '')

    try:
        print info(charm_id)
    except Exception as e:
        print e.strerror


if __name__ == '__main__':
    main()



import os
import sys
import glob


def parser_defaults(parser):
    parser.add_argument('-b', '--bundle', action='store_true',
                        help='Process as a bundle')
    parser.add_argument('--debug', action='store_true',
                        help='Provide additional debug information')

    return parser


def usage(exit_code=0, bundle=False):
    sys.stderr.write('usage: %s subcommand\n' % os.path.basename(sys.argv[0]))
    subs = subcommands(os.path.dirname(os.path.realpath(__file__)))
    sys.stderr.write('\n  Available subcommands are:\n    ')
    sys.stderr.write('\n    '.join(subs))
    sys.stderr.write('\n')
    sys.exit(exit_code)


def subcommands(scripts_dir):
    subs = []
    for path in os.environ['PATH'].split(os.pathsep):
        path = path.strip('"')
        for cmd in glob.glob(os.path.join(path, 'charm-*%s' % ext())):
            sub = os.path.basename(cmd)
            sub = sub.split('charm-')[1].replace(ext(), '')
            subs.append(sub)

    subs = sorted(set(subs))
    # Removes blacklisted items from the subcommands list.
    return filter(lambda s: s not in ['mr', 'charms'], subs)


def ext():
    return '.exe' if os.name == 'nt' else ''



#!/usr/bin/env python
#
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).

import ez_setup


ez_setup.use_setuptools()

from setuptools import setup


setup(
    name='charm-tools',
    version="1.2.8",
    packages=['charmtools'],
    install_requires=['launchpadlib', 'argparse', 'cheetah', 'pyyaml',
                      'pycrypto', 'paramiko', 'bzr', 'requests', 'markdown',
                      'charmworldlib'],
    package_data={'charmtools': ['templates/*/*.*', 'templates/*/hooks/*']},
    maintainer='Marco Ceppi',
    maintainer_email='marco@ceppi.net',
    description=('Tools for maintaining Juju charms'),
    license='GPL v3',
    url='https://launchpad.net/charm-tools',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
    ],
    entry_points={
        'console_scripts': [
            'charm = charmtools:charm',
            'juju-charm = charmtools:charm',
            'juju-bundle = charmtools:bundle',
            'juju-test = charmtools.test:main',
            'charm-get = charmtools.get:main',
            'charm-getall = charmtools.getall:main',
            'charm-proof = charmtools.proof:main',
            'charm-create = charmtools.create:main',
            'charm-list = charmtools.list:main',
            'charm-promulgate = charmtools.promulgate:main',
            'charm-review = charmtools.review:main',
            'charm-review-queue = charmtools.review_queue:main',
            'charm-search = charmtools.search:main',
            'charm-subscribers = charmtools.subscribers:main',
            'charm-unpromulgate = charmtools.unpromulgate:main',
            'charm-update = charmtools.update:main',
            'charm-version = charmtools.version:main',
            'charm-help = charmtools.cli:usage',
            'charm-test = charmtools.test:main',
            'charm-info = charmtools.info:main',
            'charm-generate = charmtools.generate:main',
            'charm-add = charmtools.generate:main',
        ],
    },
)



#!/usr/bin/env python
# Copyright (C) 2013 Marco Ceppi <marco@ceppi.net>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import argparse
import subprocess

from .mr import Mr
from . import ext


def setup_parser():
    parser = argparse.ArgumentParser(prog='juju charm getall',
                                     description='Retrieves all charms from '
                                                 'Launchpad')
    parser.add_argument('charms_directory', nargs='?',
                        help='Path to where all charms should be downloaded')

    return parser


def main():
    parser = setup_parser()
    args = parser.parse_args()

    if not args.charms_directory:
        sys.stderr.write('Error: No value for charms_directory provided\n\n')
        parser.print_help()
        sys.exit(1)

    if not os.path.exists(args.charms_directory):
        os.makedirs(args.charms_directory, 0o755)

    update_cmd = os.path.join(os.path.dirname(os.path.realpath(sys.argv[0])),
                              'charm-update%s' % ext)
    charm_update = subprocess.call([update_cmd, args.charms_directory])
    if charm_update != 0:
        sys.stderr.write('Unable to perform `juju charm update`!\n')
        sys.exit(1)

    try:
        mr = Mr(directory=args.charms_directory)
        sys.stderr.write('Grabbing %s charms from Charm Store\n' %
                         len(mr.list()))
        for charm in mr.list():
            sys.stderr.write('Branching %s\n' % charm)
            try:
                mr.update(charm)
            except (KeyboardInterrupt, SystemExit):
                sys.stderr.write('\nKeyboard Interrupt caught. Exiting!\n')
                break
            except Exception as e:
                print >> sys.stderr, "Error during update: ", e
    except Exception as e:
        print >> sys.stderr, "Error during setup: ", e





class Linter(object):
    def __init__(self, debug=False):
        self.lint = []
        self.exit_code = 0
        self.debug = debug

    def crit(self, msg):
        """Called when checking cannot continue."""
        self.err("FATAL: " + msg)

    def err(self, msg):
        global EXIT_CODE
        self.lint.append("E: " + msg)
        if self.exit_code < 200:
            self.exit_code = 200

    def info(self, msg):
        """Ignorable but sometimes useful."""
        self.lint.append("I: " + msg)

    def warn(self, msg):
        global EXIT_CODE
        self.lint.append("W: " + msg)
        if self.exit_code < 100:
            self.exit_code = 100



import os

import re
import yaml
import hashlib
import email.utils

from stat import ST_MODE
from stat import S_IXUSR

from linter import Linter
from launchpadlib.launchpad import Launchpad

KNOWN_METADATA_KEYS = ['name',
                       'summary',
                       'maintainer',
                       'description',
                       'categories',
                       'subordinate',
                       'provides',
                       'requires',
                       'format',
                       'peers']

KNOWN_RELATION_KEYS = ['interface', 'scope', 'limit', 'optional']

KNOWN_SCOPES = ['global', 'container']

TEMPLATE_PATH = os.path.abspath(os.path.dirname(__file__))

TEMPLATE_README = os.path.join(TEMPLATE_PATH, 'templates', 'charm',
                               'README.ex')

TEMPLATE_ICON = os.path.join(TEMPLATE_PATH, 'templates', 'charm', 'icon.svg')

KNOWN_OPTION_KEYS = set(('description', 'type', 'default'))

KNOWN_OPTION_TYPES = {
    'string': basestring,
    'int': int,
    'float': float,
    'boolean': bool,
}


class RelationError(Exception):
    pass


class CharmLinter(Linter):
    def check_hook(self, hook, hooks_path, recommended=False):
        hook_path = os.path.join(hooks_path, hook)

        try:
            mode = os.stat(hook_path)[ST_MODE]
            if not mode & S_IXUSR:
                self.info(hook + " not executable")
            with open(hook_path, 'r') as hook_file:
                count = 0
                for line in hook_file:
                    count += 1
                    hook_warnings = [
                        {'re': re.compile("http://169\.254\.169\.254/"),
                         'msg': "hook accesses EC2 metadata service directly"}]
                    for warning in hook_warnings:
                        if warning['re'].search(line):
                            self.warn(
                                "(%s:%d) - %s" %
                                (hook, count, warning['msg']))
            return True

        except OSError:
            if recommended:
                self.info("missing recommended hook " + hook)
            return False

    def check_relation_hooks(self, relations, subordinate, hooks_path):
        template_interfaces = ('interface-name')
        template_relations = ('relation-name')

        for r in relations.items():
            if r[0].startswith('juju-'):
                self.info('juju-* is a reserved relation name')
            if type(r[1]) != dict:
                self.err("relation %s is not a map" % (r[0]))
            else:
                if 'scope' in r[1]:
                    scope = r[1]['scope']
                    if scope not in KNOWN_SCOPES:
                        self.err("Unknown scope found in relation %s - (%s)" %
                                 (r[0], scope))
                if 'interface' in r[1]:
                    interface = r[1]['interface']
                    if interface in template_interfaces:
                        self.err("template interface names should be "
                                 "changed: " + interface)
                else:
                    self.err("relation missing interface")
                for key in r[1].keys():
                    if key not in KNOWN_RELATION_KEYS:
                        self.err(
                            "Unknown relation field in relation %s - (%s)" %
                            (r[0], key))

            r = r[0]

            if r in template_relations:
                self.err("template relations should be renamed to fit "
                         "charm: " + r)

            has_one = False
            has_one = has_one or self.check_hook(
                r + '-relation-changed', hooks_path)
            has_one = has_one or self.check_hook(
                r + '-relation-departed', hooks_path)
            has_one = has_one or self.check_hook(
                r + '-relation-joined', hooks_path)
            has_one = has_one or self.check_hook(
                r + '-relation-broken', hooks_path)

            if not has_one and not subordinate:
                self.info("relation " + r + " has no hooks")

    def check_config_file(self, charm_path):
        config_path = os.path.join(charm_path, 'config.yaml')
        if not os.path.isfile(config_path):
            self.info('File config.yaml not found.')
            return
        try:
            with open(config_path) as config_file:
                config = yaml.safe_load(config_file.read())
        except Exception, error:
            self.err('Cannot parse config.yaml: %s' % error)
            return
        if not isinstance(config, dict):
            self.err('config.yaml not parsed into a dictionary.')
            return
        if 'options' not in config:
            self.err('config.yaml must have an "options" key.')
            return
        if len(config) > 1:
            wrong_keys = sorted(config)
            wrong_keys.pop(wrong_keys.index('options'))
            self.warn('Ignored keys in config.yaml: %s' % wrong_keys)

        options = config['options']
        if not isinstance(options, dict):
            self.err(
                'config.yaml: options section is not parsed as a dictionary')
            return

        for option_name, option_value in options.items():
            if not re.match('^[a-z0-9]+[\w-]+[a-z0-9]+$', option_name,
                            flags=re.IGNORECASE):
                self.err('config.yaml: %s does not conform to naming pattern'
                         % option_name)
            if not isinstance(option_value, dict):
                self.err(
                    'config.yaml: data for option %s is not a dict'
                    % option_name)
                continue
            existing_keys = set(option_value)
            missing_keys = KNOWN_OPTION_KEYS - existing_keys
            if missing_keys:
                self.warn(
                    'config.yaml: option %s does not have the keys: %s' % (
                        option_name, ', '.join(sorted(missing_keys))))
            invalid_keys = existing_keys - KNOWN_OPTION_KEYS
            if invalid_keys:
                invalid_keys = [str(key) for key in sorted(invalid_keys)]
                self.warn(
                    'config.yaml: option %s has unknown keys: %s' % (
                        option_name, ', '.join(invalid_keys)))

            if 'description' in existing_keys:
                if not isinstance(option_value['description'], basestring):
                    self.warn(
                        'config.yaml: description of option %s should be a '
                        'string' % option_name)
            option_type = option_value.get('type', 'string')
            if option_type not in KNOWN_OPTION_TYPES:
                self.warn('config.yaml: option %s has an invalid type (%s)'
                          % (option_name, option_type))
            elif 'default' in option_value:
                expected_type = KNOWN_OPTION_TYPES[option_value['type']]
                if not isinstance(option_value['default'], expected_type):
                    self.err(
                        'config.yaml: type of option %s is specified as '
                        '%s, but the type of the default value is %s'
                        % (option_name, option_value['type'],
                           type(option_value['default']).__name__))
            else:
                # Nothing to do: the option type is valid but no default
                # value exists.
                pass


class Charm(object):
    def __init__(self, path):
        self.charm_path = path
        if not self.is_charm():
            raise Exception('Not a Charm')

    def is_charm(self):
        return os.path.isfile(os.path.join(self.charm_path, 'metadata.yaml'))

    def proof(self, remote=True):
        lint = CharmLinter()
        charm_name = self.charm_path
        if os.path.isdir(charm_name):
            charm_path = charm_name
        else:
            charm_home = os.getenv('CHARM_HOME', '.')
            charm_path = os.path.join(charm_home, charm_name)

        if not os.path.isdir(charm_path):
            lint.crit("%s is not a directory, Aborting" % charm_path)
            return lint.lint, lint.exit_code

        hooks_path = os.path.join(charm_path, 'hooks')
        yaml_path = os.path.join(charm_path, 'metadata.yaml')
        try:
            yamlfile = open(yaml_path, 'r')
            try:
                charm = yaml.safe_load(yamlfile)
            except Exception as e:
                lint.crit('cannot parse ' + yaml_path + ":" + str(e))
                return lint.lint, lint.exit_code

            yamlfile.close()

            for key in charm.keys():
                if key not in KNOWN_METADATA_KEYS:
                    lint.err("Unknown root metadata field (%s)" % key)

            charm_basename = os.path.basename(charm_path)
            if charm['name'] != charm_basename:
                warn_msg = ("metadata name (%s) must match directory name (%s)"
                            " exactly for local deployment.") % (
                                charm['name'], charm_basename)
                lint.warn(warn_msg)

            # summary should be short
            if len(charm['summary']) > 72:
                lint.warn('summary sould be less than 72')

            # need a maintainer field
            if 'maintainer' not in charm:
                lint.err('Charms need a maintainer (See RFC2822) - '
                         'Name <email>')
            else:
                if type(charm['maintainer']) == list:  # It's a list
                    maintainers = charm['maintainer']
                else:
                    maintainers = [charm['maintainer']]
                for maintainer in maintainers:
                    (name, address) = email.utils.parseaddr(maintainer)
                    formatted = email.utils.formataddr((name, address))
                    if formatted != maintainer:
                        warn_msg = ("Maintainer address should contain a "
                                    "real-name and email only. [%s]" % (
                                        formatted))
                        lint.warn(warn_msg)

            if 'categories' not in charm:
                lint.warn('Metadata is missing categories.')
            else:
                categories = charm['categories']
                if type(categories) != list or categories == []:
                    # The category names are not validated because they may
                    # change.
                    lint.warn(
                        'Categories metadata must be a list of one or more of:'
                        ' applications, app-servers, databases, file-servers, '
                        'cache-proxy, misc')

            if not os.path.exists(os.path.join(charm_path, 'icon.svg')):
                lint.warn("No icon.svg file.")
            else:
                # should have an icon.svg
                template_sha1 = hashlib.sha1()
                icon_sha1 = hashlib.sha1()
                try:
                    with open(TEMPLATE_ICON) as ti:
                        template_sha1.update(ti.read())
                        with open(os.path.join(charm_path, 'icon.svg')) as ci:
                            icon_sha1.update(ci.read())
                    if template_sha1.hexdigest() == icon_sha1.hexdigest():
                        lint.warn("Includes template icon.svg file.")
                except IOError as e:
                    lint.warn(
                        "Error while opening %s (%s)" %
                        (e.filename, e.strerror))

            # Must have a hooks dir
            if not os.path.exists(hooks_path):
                lint.info("no hooks directory")

            # Must have a copyright file
            if not os.path.exists(os.path.join(charm_path, 'copyright')):
                lint.warn("no copyright file")

            # should have a readme
            root_files = os.listdir(charm_path)
            found_readmes = set()
            for filename in root_files:
                if filename.upper().find('README') != -1:
                    found_readmes.add(filename)
            if len(found_readmes):
                if 'README.ex' in found_readmes:
                    lint.warn("Includes template README.ex file")
                try:
                    with open(TEMPLATE_README) as tr:
                        bad_lines = []
                        for line in tr:
                            if len(line) >= 40:
                                bad_lines.append(line.strip())
                        for readme in found_readmes:
                            readme_path = os.path.join(charm_path, readme)
                            with open(readme_path) as r:
                                readme_content = r.read()
                                lc = 0
                                for l in bad_lines:
                                    if not len(l):
                                        continue
                                    lc += 1
                                    if l in readme_content:
                                        err_msg = ('%s Includes line %d of '
                                                   'boilerplate README.ex ')
                                        lint.warn(err_msg % (readme, lc))
                except IOError as e:
                    lint.warn(
                        "Error while opening %s (%s)" %
                        (e.filename, e.strerror))
            else:
                lint.warn("no README file")

            subordinate = charm.get('subordinate', False)
            if type(subordinate) != bool:
                lint.err("subordinate must be a boolean value")

            # All charms should provide at least one thing
            provides = charm.get('provides')
            if provides is not None:
                lint.check_relation_hooks(provides, subordinate, hooks_path)
            else:
                if not subordinate:
                    lint.info("all charms should provide at least one thing")

            if subordinate:
                try:
                    requires = charm.get('requires')
                    if requires is not None:
                        found_scope_container = False
                        for rel_name, rel in requires.iteritems():
                            if 'scope' in rel:
                                if rel['scope'] == 'container':
                                    found_scope_container = True
                                    break
                        if not found_scope_container:
                            raise RelationError
                    else:
                        raise RelationError
                except RelationError:
                    lint.err("subordinates must have at least one scope: "
                             "container relation")
            else:
                requires = charm.get('requires')
                if requires is not None:
                    lint.check_relation_hooks(requires, subordinate,
                                              hooks_path)

            peers = charm.get('peers')
            if peers is not None:
                lint.check_relation_hooks(peers, subordinate, hooks_path)

            if 'revision' in charm:
                lint.warn("Revision should not be stored in metadata.yaml "
                          "anymore. Move it to the revision file")
                # revision must be an integer
                try:
                    x = int(charm['revision'])
                    if x < 0:
                        raise ValueError
                except (TypeError, ValueError):
                    lint.warn("revision should be a positive integer")

            lint.check_hook('install', hooks_path, recommended=True)
            lint.check_hook('start', hooks_path, recommended=True)
            lint.check_hook('stop', hooks_path, recommended=True)
            if os.path.exists(os.path.join(charm_path, 'config.yaml')):
                lint.check_hook('config-changed', hooks_path, recommended=True)
            else:
                lint.check_hook('config-changed', hooks_path)
        except IOError:
            lint.err("could not find metadata file for " + charm_name)
            lint.exit_code = -1

        # Should not have autogen test
        if os.path.exists(os.path.join(charm_path, 'tests', '00-autogen')):
            lint.warn('Includes template test file, tests/00-autogen')

        rev_path = os.path.join(charm_path, 'revision')
        if os.path.exists(rev_path):
            with open(rev_path, 'r') as rev_file:
                content = rev_file.read().rstrip()
                try:
                    int(content)
                except ValueError:
                    lint.err("revision file contains non-numeric data")

        lint.check_config_file(charm_path)
        return lint.lint, lint.exit_code

    def metadata(self):
        metadata = None
        with open(os.path.join(self.charm_path, 'metadata.yaml')) as f:
            metadata = yaml.safe_load(f.read())

        return metadata

    def promulgate(self):
        pass


def remote():
    lp = Launchpad.login_anonymously('charm-tools', 'production',
                                     version='devel')
    charm = lp.distributions['charms']
    current_series = str(charm.current_series).split('/').pop()
    branches = charm.getBranchTips()
    charms = []

    for branch in branches:
        try:
            branch_series = str(branch[2][0]).split('/')[0]
            charm_name = str(branch[0]).split('/')[3]
        except IndexError:
            branch_series = ''
        if branch_series == current_series:
            charms.append("lp:charms/%s" % charm_name)
        else:
            charms.append("lp:%s" % branch[0])
    return charms


def local(directory):
    '''Show charms that actually exist locally. Different than Mr.list'''
    local_charms = []
    for charm in os.listdir(directory):
        if os.path.exists(os.join(directory, charm, '.bzr')):
            local_charms.append(charm)
    return local_charms
